@@ -31,7 +31,7 @@ use POSIX qw(locale_h);
 setlocale(LC_NUMERIC,"C");
 
 
-$VERSION = '8.8';
+$VERSION = '8.9';
 
 # Cost scores used when converting PLSQL to PLPGSQL
 my %size_score = (
@@ -62,6 +62,7 @@ my %uncovered_score = (
 	'UTL_' => 2,
 	'EXTRACT' => 3,
 	'EXCEPTION' => 2,
+	'SUBSTR' => 1,
 );
 
 =head1 NAME
@@ -87,7 +88,7 @@ This function return a PLSQL code translated to PLPGSQL code
 
 sub plsql_to_plpgsql
 {
-        my ($str, $pg_numeric_type, $default_numeric) = @_;
+        my ($str, $allow_code_break) = @_;
 
 	#--------------------------------------------
 	# PL/SQL to PL/PGSQL code conversion
@@ -129,8 +130,28 @@ sub plsql_to_plpgsql
 
 	# dup_val_on_index => unique_violation : already exist exception
 	$str =~ s/\bdup_val_on_index\b/unique_violation/igs;
+
 	# Replace raise_application_error by PG standard RAISE EXCEPTION
-	$str =~ s/\braise_application_error[\s\t]*\([\s\t]*[\-\+]*\d+[\s\t]*,[\s\t]*(.*)\);/RAISE EXCEPTION $1;/igs;
+	$str =~ s/\braise_application_error[\s\t]*\([\s\t]*[\-\+]*\d+[\s\t]*,[\s\t]*(.*?)\);/RAISE EXCEPTION $1;/igs;
+	# and then rewrite RAISE EXCEPTION concatenations
+	while ($str =~ /RAISE EXCEPTION[\s\t]*([^;\|]+?)(\|\|)([^;]*);/) {
+		my @ctt = split(/\|\|/, "$1$2$3");
+		my $sbt = '';
+		my @args = '';
+		for (my $i = 0; $i <= $#ctt; $i++) {
+			if (($ctt[$i] =~ s/^[\s\t]*'//s) && ($ctt[$i] =~ s/'[\s\t]*$//s)) {
+				$sbt .= "$ctt[$i]";
+			} else {
+				$sbt .= '%';
+				push(@args, $ctt[$i]);
+			}
+		}
+		$sbt = "'$sbt'";
+		if ($#args >= 0) {
+			$sbt = $sbt . join(',', @args);
+		}
+		$str =~ s/RAISE EXCEPTION[\s\t]*([^;\|]+?)(\|\|)([^;]*);/RAISE EXCEPTION $sbt;/is
+	};
 
 	# Remove IN information from cursor declaration
 	while ($str =~ s/(\bCURSOR\b[^\(]+)\(([^\)]+\bIN\b[^\)]+)\)/$1\(\%\%CURSORREPLACE\%\%\)/is) {
@@ -142,8 +163,6 @@ sub plsql_to_plpgsql
 	# Reorder cursor declaration
 	$str =~ s/\bCURSOR\b[\s\t]*([A-Z0-9_\$]+)/$1 CURSOR/isg;
 
-	# Rewrite RAISE EXCEPTION concatenations
-	while ($str =~ s/RAISE EXCEPTION([^;]+)'[\s\t]*\|\|[\s\t]*([^\|]+)[\s\t]*\|\|[\s\t]*'([^;]+);/RAISE EXCEPTION$1\%$3,$2;/is) {};
 	# Rewrite TO_DATE formating call
 	$str =~ s/TO_DATE[\s\t]*\([\s\t]*('[^\']+'),[\s\t]*('[^\']+')[^\)]*\)/to_date($1,$2)/igs;
 
@@ -209,7 +228,18 @@ sub plsql_to_plpgsql
 	$str =~ s/\bSTORAGE_ERROR\b/OUT_OF_MEMORY/igs;
 	# PROGRAM_ERROR => INTERNAL ERROR ?
 	# ROWTYPE_MISMATCH => DATATYPE MISMATCH ?
-	
+
+	if ($allow_code_break) {
+		# Replace Oracle substr(string, start_position, length) with
+		# PostgreSQL substring(string from start_position for length)
+		$str =~ s/substr[\s\t]*\(([^;]*),[\s\t]*(\d+)[\s\t]*,[\s\t]*(\d+)[\s\t]*\)/substring($1 from $2 for $3)/igs;
+		$str =~ s/substr[\s\t]*\(([^;]*),[\s\t]*(\d+)[\s\t]*\)/substring($1 from $2)/igs;
+
+		# Replace decode("user_status",'active',"username",null)
+		# PostgreSQL (CASE WHEN "user_status"='ACTIVE' THEN "username" ELSE NULL END)
+		$str =~ s/decode[\s\t]*\([\s\t]*([^,]*),[\s\t]*([^,]*),[\s\t]*([^,]*),[\s\t]*([^\)]*)\)/\(CASE WHEN $1=$2 THEN $3 ELSE $4 END\)/igs;
+	}
+
 	return $str;
 }
 
@@ -390,6 +420,8 @@ sub estimate_cost
 	$cost += $uncovered_score{'UTL_'}*$n;
 	$n = () = $str =~ m/\bEXTRACT[\t\s]*\(/igs;
 	$cost += $uncovered_score{'EXTRACT'}*$n;
+	$n = () = $str =~ m/\bSUBSTR[\t\s]*\(/igs;
+	$cost += $uncovered_score{'SUBSTR'}*$n;
 	# See:  http://www.postgresql.org/docs/9.0/static/errcodes-appendix.html#ERRCODES-TABLE
 	$n = () = $str =~ m/\b(DUP_VAL_ON_INDEX|TIMEOUT_ON_RESOURCE|TRANSACTION_BACKED_OUT|NOT_LOGGED_ON|LOGIN_DENIED|INVALID_NUMBER|PROGRAM_ERROR|VALUE_ERROR|ROWTYPE_MISMATCH|CURSOR_ALREADY_OPEN|ACCESS_INTO_NULL|COLLECTION_IS_NULL)\b/igs;
 	$cost += $uncovered_score{'EXCEPTION'}*$n;
